/*
 * Test Mesa (OpenGL) drawing primitives.
 *
 * Copyright (C) 2002 Leandro Bernsmller
 * License: GNU-General Public License 2
 *
 * Based on Jeff Molofee code found at http://nehe.gamedev.net 
 * (which was ported to Linux/GLUT by Richard Campbell).
 *
 * Some parts of the code are commented,
 * because they are from the original code
 * and will be used on next steps.
 */



#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>



// some definitions
// keys
#define KEY_ESC 27
// default y Field Of View (FOV - the viewer vertical angle of vision)
#define INI_YFOV 45.0f
// default output window size
#define INI_WIDTH 320
#define INI_HEIGHT 240



// program output window ID
int window_id; 



/*
 * Handle window resize messages.
 * w - output window width (in pixels)
 * h - output window height (in pixels)
 */
void handle_resize (int w, int h)
{
  // if input is zero, make it 1 to avoid division by zero after
  if (h == 0) h = 1;
  if (w == 0) w = 1;

  // define y angle (field of view)
  GLfloat yfov = INI_YFOV;
  // if width is proportionaly less than originaly
  if ( (GLfloat) w / h < (GLfloat) INI_WIDTH / INI_HEIGHT ) {

    // Enlarge y field of view to reduce drawing size (all things visible in original
    // window size, continue visible in all window sizes).
    // define minimal width for current height (less than this needs fov correction)
    float min_width = INI_WIDTH * h / INI_HEIGHT;
    // correct fov
    yfov = INI_YFOV * min_width / w;
    // to avoid desapear and fails, stop enlarge with this value
    if ( yfov > 160 ) yfov = 160;
  }

  glViewport (0, 0, w, h);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (yfov, (GLfloat) w / (GLfloat) h, 0.1f, 100.0f);
  glMatrixMode (GL_MODELVIEW);
}



/* 
 * Draw the scene.
 */
void handle_draw_scene ()
{
  glClear (GL_COLOR_BUFFER_BIT);
  //glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity ();

  // using perpective view, need move forward or you will be to close with objects
  glTranslatef (-1.5f, 0.0f, -6.0f);  // Move Left 1.5 Units And Into The Screen 6.0
	
  // draw a triangle
  glBegin (GL_POLYGON);
  glVertex3f ( 0.0f, 1.0f, 0.0f);   // Top
  glVertex3f ( 1.0f,-1.0f, 0.0f);   // Bottom Right
  glVertex3f (-1.0f,-1.0f, 0.0f);   // Bottom Left	
  glEnd ();

  glTranslatef (3.0f,0.0f,0.0f);    // Move Right 3 Units
	
  // draw a square (quadrilateral)
  glBegin (GL_QUADS);
  glVertex3f (-1.0f, 1.0f, 0.0f);   // Top Left
  glVertex3f ( 1.0f, 1.0f, 0.0f);   // Top Right
  glVertex3f ( 1.0f,-1.0f, 0.0f);   // Bottom Right
  glVertex3f (-1.0f,-1.0f, 0.0f);   // Bottom Left	
  glEnd ();
  
  // draw a box around the quadrilateral, over plane z=0
  glBegin (GL_LINE_STRIP);
  glVertex2f (-1.2f, 1.2f);   // Top Left
  glVertex2f ( 1.2f, 1.2f);   // Top Right
  glVertex2f ( 1.2f,-1.2f);   // Bottom Right
  glVertex2f (-1.2f,-1.2f);   // Bottom Left	
  glVertex2f (-1.2f, 1.2f);   // Top Left (to close box)
  glEnd ();

  // swap buffers to display, since we're double buffered.
  //glutSwapBuffers();

  // verify if happens something wrong
  GLenum err = glGetError ();
  if (err != GL_NO_ERROR) {
    printf ("Error while using OpenGL stuff.\nAborting...\n");
    exit (1);
  }
}



/* 
 * Called when a key is pressed.
 * key - ASCII code for the key
 * x, y - window coordenates of mouse when key was pressed
 */
void handle_keyboard (unsigned char key, int x, int y) 
{
  // if escape is pressed, end the program
  if (key == KEY_ESC) { 

    // close the window
    glutDestroyWindow (window_id); 

    // end
    exit(0);                   
  }
}



int main(int argc, char **argv) 
{  
  // prepare OpenGL
  glutInit (&argc, argv);  
  //glutInitDisplayMode (GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);  

  // define and create the output window (viewport)
  int w = INI_WIDTH;
  int h = INI_HEIGHT;
  glutInitWindowSize (w, h);
  glutInitWindowPosition (0, 0);  
  window_id = glutCreateWindow ("Test Drawing Primitives");  

  // glutFullScreen();

  // define functions to update application state acording events, etc.
  // define the user render function
  glutDisplayFunc (&handle_draw_scene);  
  // uncomment to: even if there are no events, redraw our gl scene
  //glutIdleFunc (&handle_draw_scene);
  // register the function called when our output window is resized.
  glutReshapeFunc (&handle_resize);
  // register the function called when the keyboard is pressed.
  glutKeyboardFunc (&handle_keyboard);

  // prepare rendering buffers
  glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
  glClearDepth (1.0);
  glDisable (GL_DEPTH_TEST);
  // glShadeModel (GL_SMOOTH); // Enables Smooth Color Shading

  // setup universe window to output window transformations
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (INI_YFOV, (GLfloat) w / (GLfloat) h, 0.1f, 100.0f);
  glMatrixMode (GL_MODELVIEW);

  glutMainLoop();  

  return 0;
}
