/*
 * Draw one vector and its interseccion with one square area limit.
 * User can rotate the vector pressing the keys 'z' and 'x'.
 *
 * Copyright (C) 2002 Leandro Bernsmller
 * License: GNU-General Public License 2
 */



#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>
#include <math.h>



// some definitions
// keys
#define KEY_ESC 27
// default y Field Of View (FOV - the viewer vertical angle of vision)
#define INI_YFOV 45.0f
// default output window size
#define INI_WIDTH 320
#define INI_HEIGHT 240
// define PI, if need
#ifndef PI
#define PI 3.141592
#endif



// global variables
// program output window ID
int window_id;
float view_angle = 0;




/*
 * Handle window resize messages.
 * w - output window width (in pixels)
 * h - output window height (in pixels)
 */
void handle_resize (int w, int h)
{
  // if input is zero, make it 1 to avoid division by zero after
  if (h == 0) h = 1;
  if (w == 0) w = 1;

  // define y angle (field of view)
  GLfloat yfov = INI_YFOV;
  // if width is proportionaly less than originaly
  if ( (GLfloat) w / h < (GLfloat) INI_WIDTH / INI_HEIGHT ) {

    // Enlarge y field of view to reduce drawing size (all things visible in original
    // window size, continue visible in all window sizes).
    // define minimal width for current height (less than this needs fov correction)
    float min_width = INI_WIDTH * h / INI_HEIGHT;
    // correct fov
    yfov = INI_YFOV * min_width / w;
    // to avoid desapear and fails, stop enlarge with this value
    if ( yfov > 160 ) yfov = 160;
  }

  glViewport (0, 0, w, h);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (yfov, (GLfloat) w / (GLfloat) h, 0.1f, 300.0f);
  glMatrixMode (GL_MODELVIEW);
}



/* 
 * Draw the scene.
 */
void handle_draw_scene ()
{
  glClear (GL_COLOR_BUFFER_BIT);
  //glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity ();

  // using perpective view, need move forward or you will be to close with objects
  glTranslatef ( 0.0, 0.0, -250.0); // Move Into The Screen
  glScalef (1, -1, 1);              // Flip Vertically The Map

  // square dimensions (must be the same)
  GLint h = 100;
  GLint l = 100;

  // setup colors
  glColor3f (0.5, 0.5, 0.5);
	
  // draw the coordenate axes
  glBegin (GL_LINES);
  glVertex2i ( 0, h);
  glVertex2i ( 0,-h);
  glVertex2i (-l, 0);
  glVertex2i ( l, 0);
  glEnd ();
  
  // draw the area limit with a box
  glBegin (GL_LINE_STRIP);
  glVertex2i (-l/2, h/2);   // Top Left
  glVertex2i ( l/2, h/2);   // Top Right
  glVertex2i ( l/2,-h/2);   // Bottom Right
  glVertex2i (-l/2,-h/2);   // Bottom Left	
  glVertex2i (-l/2, h/2);   // Top Left (to close box)
  glEnd ();

  // define and draw one vector

  // define vector color
  glColor3f (1, 1, 1);

  // vector components
  GLfloat vx = 45;
  GLfloat vy = 0;

  // rotate vector
  GLfloat rd = view_angle * PI / 180;
  GLfloat cs = cos (rd);
  GLfloat sn = sin (rd);
  GLfloat tmp_x = vx;
  GLfloat tmp_y = vy;
  vx = tmp_x * cs + tmp_y * sn;
  vy = tmp_x * sn + tmp_y * cs;

  // draw vector
  glBegin (GL_LINES);
  glVertex2i ( 0, 0);
  glVertex2i ( (GLint) vx, (GLint) vy);
  glEnd ();

  // define interseccion coordenate between vector line and area limits
  GLfloat ix;
  GLfloat iy;
  if (view_angle <= 45 || view_angle >= 315) {
    ix = l / 2;
    iy = h / 2 * sn / cs;
  }
  if (view_angle <= 225 && view_angle >= 135) {
    ix = -l / 2;
    iy = -h / 2 * sn / cs;
  }
  if (view_angle > 45 && view_angle < 135) {
    ix = l / 2 * cs / sn;
    iy = h / 2;
  }
  if (view_angle > 225 && view_angle < 315) {
    ix = -l / 2 * cs / sn;
    iy = -h / 2;
  }

  // draw one cross in the interseccion point
  GLint cx = (GLint) ix;
  GLint cy = (GLint) iy;
  glBegin (GL_LINES);
  glVertex2i ( cx - 2, cy - 2);
  glVertex2i ( cx + 2, cy + 2);
  glEnd ();
  glBegin (GL_LINES);
  glVertex2i ( cx + 2, cy - 2);
  glVertex2i ( cx - 2, cy + 2);
  glEnd ();

  // swap buffers to display, since we're double buffered.
  //glutSwapBuffers();

  // verify if happens something wrong
  GLenum err = glGetError ();
  if (err != GL_NO_ERROR) {
    printf ("Error while using OpenGL stuff.\nAborting...\n");
    exit (1);
  }
}



/* 
 * Called when a key is pressed.
 * key - ASCII code for the key
 * x, y - window coordenates of mouse when key was pressed
 */
void handle_keyboard (unsigned char key, int x, int y) 
{
  char redraw = 0;   // flag if need redraw the scene

  // if escape is pressed, end the program
  switch (key) {

  case KEY_ESC:   // close the window, end
    glutDestroyWindow (window_id);
    exit (0);
    break;

  case 'z':   // decrease angle
  case 'Z':
    view_angle -= 5;
    redraw = 1;
    break;

  case 'x':   // increase angle
  case 'X':
    view_angle += 5;
    redraw = 1;
    break;
  }

  // mantain angle between 0 and 360 (excluding 360)
  if (view_angle < 0) view_angle = 360 + view_angle;
  if (view_angle >= 360) view_angle = view_angle - 360;

  if (redraw) handle_draw_scene ();
}



int main(int argc, char **argv) 
{  
  // prepare OpenGL
  glutInit (&argc, argv);  
  //glutInitDisplayMode (GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);  

  // define and create the output window (viewport)
  int w = INI_WIDTH;
  int h = INI_HEIGHT;
  glutInitWindowSize (w, h);
  glutInitWindowPosition (0, 0);  
  window_id = glutCreateWindow ("Interseccion With Limit");  

  // glutFullScreen();

  // define functions to update application state acording events, etc.
  // define the user render function
  glutDisplayFunc (&handle_draw_scene);  
  // uncomment to: even if there are no events, redraw our gl scene
  //glutIdleFunc (&handle_draw_scene);
  // register the function called when our output window is resized.
  glutReshapeFunc (&handle_resize);
  // register the function called when the keyboard is pressed.
  glutKeyboardFunc (&handle_keyboard);

  // prepare rendering buffers
  glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
  glClearDepth (1.0);
  glDisable (GL_DEPTH_TEST);
  // glShadeModel (GL_SMOOTH); // Enables Smooth Color Shading

  // setup universe window to output window transformations
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (INI_YFOV, (GLfloat) w / (GLfloat) h, 0.1f, 300.0f);
  glMatrixMode (GL_MODELVIEW);

  printf ("Press the keys 'z' and 'x' to rotate the vectors.\n");
  printf ("Press the key Esc to exit.\n");

  glutMainLoop();

  return 0;
}
