/*
 * Draw and rotate the hole landscape.
 * Fill the area delimited by two vectors and one square.
 * User can rotate the vectors pressing the keys 'z' and 'x'.
 * The area is filled by a custom set of colors got from a bitmap.
 *
 * Copyright (C) 2002 Leandro Bernsmller
 * License: GNU-General Public License 2
 */



#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>



// some definitions
// keys
#define KEY_ESC 27
// default y Field Of View (FOV - the viewer vertical angle of vision)
#define INI_YFOV 45.0f
// default output window size
#define INI_WIDTH 600
#define INI_HEIGHT 207
// define PI, if need
#ifndef PI
#define PI 3.141592
#endif
// the size of the side of square bitmap
#define COLORS_BMP_SIDE 100



// global variables
// program output window ID
int window_id;
// vectors rotation angle
float view_angle = 0;
// bitmap with colors to be used to fill the area
GLfloat colors_bmp [COLORS_BMP_SIDE][COLORS_BMP_SIDE];



/*
 * Define the bitmap with the colors to be used in the area fill.
 */
void define_colors_bitmap ()
{
  for (int c = 0; c < COLORS_BMP_SIDE; c++)
    for (int l = 0; l < COLORS_BMP_SIDE; l++)
      colors_bmp [c][l] = (GLfloat) (0.5 + (sin (c * PI / 12) + cos (l * PI / 12)) / 4);
}



/*
 * Return the color for the given position in colors bitmap.
 * c, l - bitmap position
 * return:
 * the color between 0 and 1
 */
GLfloat get_bitmap_color (GLint c, GLint l)
{
  // avoid positions out of bitmap range
  if (c < 0 || c >= COLORS_BMP_SIDE ||
      l < 0 || l >= COLORS_BMP_SIDE) {
    printf ("Error: try access outside color bitmap. Position: (%d, %d)\n", c, l);
    return 0;
  }

  return colors_bmp [c][l];
}



/*
 * Handle window resize messages.
 * w - output window width (in pixels)
 * h - output window height (in pixels)
 */
void handle_resize (int w, int h)
{
  // if input is zero, make it 1 to avoid division by zero after
  if (h == 0) h = 1;
  if (w == 0) w = 1;

  // define y angle (field of view)
  GLfloat yfov = INI_YFOV;
  // if width is proportionaly less than originaly
  if ( (GLfloat) w / h < (GLfloat) INI_WIDTH / INI_HEIGHT ) {

    // Enlarge y field of view to reduce drawing size (all things visible in original
    // window size, continue visible in all window sizes).
    // define minimal width for current height (less than this needs fov correction)
    float min_width = INI_WIDTH * h / INI_HEIGHT;
    // correct fov
    yfov = INI_YFOV * min_width / w;
    // to avoid desapear and fails, stop enlarge with this value
    if ( yfov > 160 ) yfov = 160;
  }

  glViewport (0, 0, w, h);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (yfov, (GLfloat) w / (GLfloat) h, 0.1f, 500.0f);
  glMatrixMode (GL_MODELVIEW);
}



/*
 * Draw the vectors and the filled area, as one landscape height map.
 */
void draw_map ()
{
  glLoadIdentity ();

  // using perpective view, need move forward or you will be to close with objects
  glTranslatef ( -220.0, 0.0, -250.0); // Move Left And Into The Screen
  glScalef (1, -1, 1);                 // Flip Vertically The Map

  // square dimensions (must be the same)
  GLint h = COLORS_BMP_SIDE;
  GLint l = COLORS_BMP_SIDE;

  // setup colors
  glColor3f (0.5, 0.5, 0.5);
	
  // draw the coordenate axes
  glBegin (GL_LINES);
  glVertex2i ( 0, h);
  glVertex2i ( 0,-h);
  glVertex2i (-l, 0);
  glVertex2i ( l, 0);
  glEnd ();
  
  // draw the area limit with a box
  glBegin (GL_LINE_STRIP);
  glVertex2i (-l/2, h/2);   // Top Left
  glVertex2i ( l/2, h/2);   // Top Right
  glVertex2i ( l/2,-h/2);   // Bottom Right
  glVertex2i (-l/2,-h/2);   // Bottom Left	
  glVertex2i (-l/2, h/2);   // Top Left (to close box)
  glEnd ();

  // define and draw vectors

  // define vectors color
  glColor3f (1, 1, 1);

  // define angle between vectors
  GLfloat vects_angle = 60;

  // vectors components
  GLfloat vx1 = 1;
  GLfloat vy1 = 0;
  GLfloat vx2 = 1;
  GLfloat vy2 = 0;

  // rotate vectors
  GLfloat a1 = view_angle + vects_angle / 2;
  GLfloat a2 = view_angle - vects_angle / 2;
  GLfloat rd1 = a1 * PI / 180;
  GLfloat rd2 = a2 * PI / 180;
  GLfloat cs1 = cos (rd1);
  GLfloat sn1 = sin (rd1);
  GLfloat cs2 = cos (rd2);
  GLfloat sn2 = sin (rd2);
  GLfloat tmp_x = vx1;
  GLfloat tmp_y = vy1;
  vx1 = tmp_x * cs1 + tmp_y * sn1;
  vy1 = tmp_x * sn1 + tmp_y * cs1;
  tmp_x = vx2;
  tmp_y = vy2;
  vx2 = tmp_x * cs2 + tmp_y * sn2;
  vy2 = tmp_x * sn2 + tmp_y * cs2;

  // slant
  // linear inclination with y in function of x
  GLfloat tg_y1 = sn1 / cs1;
  GLfloat tg_y2 = sn2 / cs2;
  // linear inclination with x in function of y
  GLfloat tg_x1 = cs1 / sn1;
  GLfloat tg_x2 = cs2 / sn2;

  // define interseccion coordenate between vectors and limits
  // for vector 1
  GLfloat ix1;
  GLfloat iy1;
  if (a1 <= 45 || a1 >= 315) {
    ix1 = l / 2;
    iy1 = h / 2 * tg_y1;
  }
  if (a1 <= 225 && a1 >= 135) {
    ix1 = -l / 2;
    iy1 = -h / 2 * tg_y1;
  }
  if (a1 > 45 && a1 < 135) {
    ix1 = l / 2 * tg_x1;
    iy1 = h / 2;
  }
  if (a1 > 225 && a1 < 315) {
    ix1 = -l / 2 * tg_x1;
    iy1 = -h / 2;
  }
  // for vector 2
  GLfloat ix2;
  GLfloat iy2;
  if (a2 <= 45 || a2 >= 315) {
    ix2 = l / 2;
    iy2 = h / 2 * tg_y2;
  }
  if (a2 <= 225 && a2 >= 135) {
    ix2 = -l / 2;
    iy2 = -h / 2 * tg_y2;
  }
  if (a2 > 45 && a2 < 135) {
    ix2 = l / 2 * tg_x2;
    iy2 = h / 2;
  }
  if (a2 > 225 && a2 < 315) {
    ix2 = -l / 2 * tg_x2;
    iy2 = -h / 2;
  }

  // draw vectors lines
  GLint cx = (GLint) ix1;
  GLint cy = (GLint) iy1;
  glBegin (GL_LINES);
  glVertex2i ( 0, 0);
  glVertex2i ( cx, cy);
  glEnd ();
  cx = (GLint) ix2;
  cy = (GLint) iy2;
  glBegin (GL_LINES);
  glVertex2i (0, 0);
  glVertex2i (cx, cy);
  glEnd ();

  // fill area
  glPointSize (1.0);    // if is 3, presents strange pan on bitmap colors
  // acording the vectors angle, is used one specific draw loop
  if (view_angle <= 45 || view_angle >= 315)
    // for earch row
    for (GLint px = 0; px < l / 2; px ++) {
      // define where begin and end on y axis
      GLint iniy = (GLint) (tg_y2 * (GLfloat) px);
      GLint endy = (GLint) (tg_y1 * (GLfloat) px);
      // clip
      if (iniy < -h / 2 + 1) iniy = -h / 2 + 1;
      if (endy > h / 2) endy = h / 2;
      // draw point by point
      glBegin (GL_POINTS);
      for (GLint py = iniy; py < endy; py ++) {
	GLfloat p_color = get_bitmap_color (px + l / 2, py + h / 2);
	glColor3f (p_color, p_color, 0);
        glVertex2i (px, py);
      }
      glEnd ();
    }
  if (view_angle >= 135 && view_angle <= 225)
    // for earch row
    for (GLint px = 0; px > -l / 2; px --) {
      // define where begin and end on y axis
      GLint iniy = (GLint) (tg_y1 * (GLfloat) px);
      GLint endy = (GLint) (tg_y2 * (GLfloat) px);
      // clip
      if (iniy < -h / 2 + 1) iniy = -h / 2 + 1;
      if (endy > h / 2) endy = h / 2;
      // draw point by point
      glBegin (GL_POINTS);
      for (GLint py = iniy; py < endy; py ++) {
	GLfloat p_color = get_bitmap_color (px + l / 2, py + h / 2);
	glColor3f (p_color, p_color, 0);
        glVertex2i (px, py);
      }
      glEnd ();
    }
  if (view_angle > 45 && view_angle < 135)
    // for earch line
    for (GLint py = 0; py < h / 2; py ++) {
      // define where begin and end on y axis
      GLint inix = (GLint) (tg_x1 * (GLfloat) py);
      GLint endx = (GLint) (tg_x2 * (GLfloat) py);
      // clip
      if (inix < -l / 2 + 1) inix = -l / 2 + 1;
      if (endx > l / 2) endx = l / 2;
      // draw point by point
      glBegin (GL_POINTS);
      for (GLint px = inix; px < endx; px ++) {
	GLfloat p_color = get_bitmap_color (px + l / 2, py + h / 2);
	glColor3f (p_color, p_color, 0);
        glVertex2i (px, py);
      }
      glEnd ();
    }
  if (view_angle > 225 && view_angle < 315)
    // for earch line
    for (GLint py = 0; py > -h / 2; py --) {
      // define where begin and end on y axis
      GLint inix = (GLint) (tg_x2 * (GLfloat) py);
      GLint endx = (GLint) (tg_x1 * (GLfloat) py);
      // clip
      if (inix < -l / 2 + 1) inix = -l / 2 + 1;
      if (endx > l / 2) endx = l / 2;
      // draw point by point
      glBegin (GL_POINTS);
      for (GLint px = inix; px < endx; px ++) {
	GLfloat p_color = get_bitmap_color (px + l / 2, py + h / 2);
	glColor3f (p_color, p_color, 0);
        glVertex2i (px, py);
      }
      glEnd ();
    }
}




/*
 * Draw the landscape.
 */
void draw_lanscape ()
{
  glColor3f (0.5, 0.5, 1);
  glLoadIdentity ();
  glTranslatef (0.0, 0.0, -300.0); // Move Into The Screen
  glRotatef (view_angle, 0, 1, 0);
  glBegin (GL_POINTS);
  for (int lc = 0; lc < COLORS_BMP_SIDE; lc++)
    for (int ll = 0; ll < COLORS_BMP_SIDE; ll++) {
      GLfloat uni_x = 4 * (lc - COLORS_BMP_SIDE / 2);
      GLfloat uni_y = -50 + 16 * get_bitmap_color (lc, ll);
      GLfloat uni_z = 4 * (ll - COLORS_BMP_SIDE / 2);
      glVertex3f (uni_x, uni_y, uni_z);
    }
  glEnd ();
}



/* 
 * Draw the scene.
 */
void handle_draw_scene ()
{
  glClear (GL_COLOR_BUFFER_BIT);
  //glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  draw_lanscape ();
  draw_map ();

  // swap buffers to display, since we're double buffered.
  glutSwapBuffers();

  // verify if happens something wrong
  GLenum err = glGetError ();
  if (err != GL_NO_ERROR) {
    printf ("Error while using OpenGL.\nAborting...\n");
    exit (1);
  }
}



/* 
 * Called when a key is pressed.
 * key - ASCII code for the key
 * x, y - window coordenates of mouse when key was pressed
 */
void handle_keyboard (unsigned char key, int x, int y) 
{
  char redraw = 0;   // flag if need redraw the scene

  // if escape is pressed, end the program
  switch (key) {

  case KEY_ESC:   // close the window, end
    glutDestroyWindow (window_id);
    exit (0);
    break;

  case 'z':   // decrease angle
  case 'Z':
    view_angle -= 5;
    redraw = 1;
    break;

  case 'x':   // increase angle
  case 'X':
    view_angle += 5;
    redraw = 1;
    break;
  }

  // mantain angle between 0 and 360 (excluding 360)
  if (view_angle < 0) view_angle = 360 + view_angle;
  if (view_angle >= 360) view_angle = view_angle - 360;

  if (redraw) handle_draw_scene ();
}



int main(int argc, char **argv) 
{  
  // prepare OpenGL
  glutInit (&argc, argv);  
  glutInitDisplayMode (GLUT_DOUBLE);  
  //glutInitDisplayMode (GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);  

  // define and create the output window (viewport)
  int w = INI_WIDTH;
  int h = INI_HEIGHT;
  glutInitWindowSize (w, h);
  glutInitWindowPosition (0, 0);  
  window_id = glutCreateWindow ("External Landscape View");  

  // glutFullScreen();

  // define functions to update application state acording events, etc.
  // define the user render function
  glutDisplayFunc (&handle_draw_scene);  
  // uncomment to: even if there are no events, redraw our gl scene
  //glutIdleFunc (&handle_draw_scene);
  // register the function called when our output window is resized.
  glutReshapeFunc (&handle_resize);
  // register the function called when the keyboard is pressed.
  glutKeyboardFunc (&handle_keyboard);

  // prepare rendering buffers
  glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
  glClearDepth (1.0);
  glDisable (GL_DEPTH_TEST);
  // glShadeModel (GL_SMOOTH); // Enables Smooth Color Shading

  // setup universe window to output window (viewport) transformations
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (INI_YFOV, (GLfloat) w / (GLfloat) h, 0.1f, 500.0f);
  glMatrixMode (GL_MODELVIEW);

  // define the colors used to fill the area
  define_colors_bitmap ();

  printf ("Press the keys 'z' and 'x' to rotate.\n");
  printf ("Press the key Esc to exit.\n");

  glutMainLoop();

  return 0;
}
